/*
        clgrep.cpp -- Container Lite v 1.89
        (C) Copyright 1995  John Webster Small
        All rights reserved
*/

#include "clgrep.h"
#include <ctype.h>

grep_tok::~grep_tok() {}

#pragma argsused
int grep_tok::matchOne
    (const char * SS, size_t& len, int caseIgnore,
    CL<char> * groupToks)
{ assert(SS && *SS); len = 1; return 1; }

grep_tok *
grep_tok::newTok
    (const char * GRE, size_t GRE_len, size_t& tokLen,
     int groups)
{
 // GRE has already been stripped of
 //   leading ^ and trailing $.
    tokLen = 0;
    assert(GRE);
    if (!GRE_len) return 0;
    if (strspn(GRE,")]|*+?|") ||
        groups && GRE[0] == '}')  {
        error(GRE,0,"illegal grep char");
        return 0;
    }
    grep_tok * T = 0;
    switch (*GRE)  {
        case '[': T = new grep_set(GRE,GRE_len,tokLen);
            break;
        case '(': T = 
            new grep_subExp(GRE,GRE_len,tokLen,groups);
            break;
        case '.': T = new grep_tok; tokLen = 1; break;
        case '{':
            if (groups)  {
                T = new grep_group(GRE,GRE_len,tokLen);
                break;
            }
            // else '{' is like any other character
        default : T = new grep_seq
                (GRE,GRE_len,tokLen,groups);
            break;
    }
    if (T) if (!tokLen)  {
        delete T;
        T = 0;
    }
    else  {
        switch (GRE[tokLen])  {
          case '*' : tokLen++; T->count = Any;
            break;
          case '+' : tokLen++; T->count = OneOrMore;
            break;
          case '?' : tokLen++; T->count = ZeroOrOne;
            break;
          case '|' :
            {
                tokLen++;
                size_t rightTokLen = 0;
                grep_tok * oldT = T;
                T = new grep_alt(&GRE[tokLen],
                    GRE_len-tokLen,T,
                    rightTokLen,groups);
                if (T)
                    tokLen += rightTokLen + 1;
                else  {
                    tokLen = 0;
                    delete oldT;
                }
            }
            break;
        }
    }
    return T;
}

void grep_tok::error(const char * GRE,
    size_t pos, const char * msg)
{
    assert(GRE);
    cerr << '\n' << GRE;
    if (pos <= strlen(GRE))  {
        cstr tmp(' ',pos);
        cerr << '\n'
             << (tmp.len()? tmp.str() : "") << '^';
    }
    if (msg)
        cerr << "  " << msg;
}

int grep_exp::parse
    (const char * GRE, size_t GRE_len, int groups)
{
    static int sema4 = 0;
    assert(sema4 < CL_GREP_MAX_RECUR);
    sema4++;
    size_t tokLen = 0;
    for (size_t i = 0; i < GRE_len; i += tokLen)
        if (!insQDL(grep_tok::
            newTok(&GRE[i],GRE_len-i,tokLen,groups)))
        {
            (void) allDel();
            sema4--;
            return 0;
        }
    sema4--;
    return 1;
}

void grep_exp::parseGroup
    (char leftDelimiter, char rightDelimiter,
    const char * GRE, size_t GRE_len, size_t& tokLen,
    int groups)
{
    static char nestcspn[] = "  \\";
    nestcspn[0] = leftDelimiter;
    nestcspn[1] = rightDelimiter;
    assert(GRE && GRE[0] == leftDelimiter);
    tokLen = 1;
    size_t cspnLen;
    cstr group;
    int nesting = 1;
    while (tokLen < GRE_len)  {
        if (!group.ncat(&GRE[tokLen],cspnLen
            = strcspn(&GRE[tokLen],nestcspn),2)
            .ok())
            break;
        tokLen += cspnLen;
        if (tokLen >= GRE_len) {
            grep_tok::error(GRE,tokLen,
                "nesting unbalanced");
            break; }
        if (GRE[tokLen] == '\\')  {
            if (tokLen + 1 >= GRE_len)  {
                grep_tok::error
                    (GRE,tokLen+1,"nothing after \\");
                break; }
            group.ncat(&GRE[tokLen],2);
            tokLen += 2;
        }
        else if (GRE[tokLen] == leftDelimiter)  {
            group.ncat(&GRE[tokLen++],1);
            nesting++;
        }
        else   {  // found ")" or "}"
            if (--nesting)
                group.ncat(&GRE[tokLen++],1);
            else  {
                tokLen++;
                if (parse(group.str(),group.len(),
                    groups))
                    return;
            }
        }
    }
    tokLen = 0;
}

int grep_exp::match(const char * SS,
    size_t& len, int caseIgnore, CL<char> * groupToks)
{
    size_t SSlen = (SS? strlen(SS) : 0);
    size_t tokLen;
    len = 0;
    grep_tok * T;
    for (unsigned i = 0; (T = atGet(i)) != 0
        && len < SSlen;
        i++, len += tokLen)  {
        if (!T->matchOne(&SS[len],tokLen,
            caseIgnore,groupToks)) {
            if (T->count == grep_tok::ZeroOrOne
                || T->count == grep_tok::Any)  {
                tokLen = 0;
                continue;
            }
            return 0; //OneOrMore, One
        }
        if (T->count == grep_tok::One
            || T->count == grep_tok::ZeroOrOne)
            continue;
        // Searching for OneOrMore, Any
        size_t maxLen = len + tokLen;
        unsigned nextTok = i + 1;
        while (maxLen < SSlen)  {
            if (nextTok < nodes())
             // Look ahead before matching more of same
                if (atGet(nextTok)->matchOne
                    (&SS[maxLen],tokLen,
                    caseIgnore,groupToks))
                    break;
            if (!T->matchOne
                (&SS[maxLen],tokLen,
                caseIgnore,groupToks))
                break;
            maxLen += tokLen;
        }
        tokLen = maxLen - len;
    }
    return 1;
}

grep_subExp::grep_subExp
    (const char * GRE, size_t GRE_len, size_t& tokLen,
    int groups)
    : grep_tok(), grep_exp()
{ parseGroup('(',')',GRE,GRE_len,tokLen,groups); }

int grep_subExp::matchOne(const char * SS,
    size_t& len, int caseIgnore, CL<char> * groupToks)
{ return grep_exp::match(SS,len,caseIgnore,groupToks); }

grep_group::grep_group
    (const char * GRE, size_t GRE_len, size_t& tokLen)
    : grep_tok(), grep_exp()
{ parseGroup('{','}',GRE,GRE_len,tokLen,1); }

int grep_group::matchOne(const char * SS,
    size_t& len, int caseIgnore, CL<char> * groupToks)
{
  if (grep_exp::match(SS,len,caseIgnore,groupToks)) {
    if (groupToks && groupToks->vacancy())  {
      char * GT = new char[len+1];
      if (GT)  {
        strncpy(GT,SS,len);
        GT[len] = '\0';
        if (!groupToks->insQDL(GT))
          (void) groupToks->setMaxNodes(CL_MAXNODES-1);
      }
      else
        groupToks->setMaxNodes(CL_MAXNODES-1);
    }
    return 1;
  }
  return 0;
}

int CL_grep::_init(const char * GRE, int groups)
{
    size_t GRE_len;
    (void) parseTree.allDel();
    gre.trunc();
    BOL_anchor = EOL_anchor = 0;
    _groups = groups;
    if (GRE && (GRE_len = strlen(GRE)) != 0)  {
        size_t i = 0;
        // look for BOL, EOL pattern anchors
        if (GRE[i] == '^')  {
            BOL_anchor = 1;
            i++;
        }
        if (GRE[GRE_len - 1] == '$')  {
            EOL_anchor = 1;
            GRE_len--;
        }
        // parse tokens
        cstr tmp;
        if (tmp.ncpy(&GRE[i],GRE_len-i).ok())
          if (parseTree.parse
            (tmp.str(),tmp.len(),groups))  {
            // GRE valid, parse tree complete
            gre.cpy(GRE);
            return 1;
          }
    }
    return 0;
}

CL_grep::CL_grep(const char * GRE, int groups)
    : parseTree(), gre()
    , BOL_anchor(0), EOL_anchor(0), _groups(0)
{ (void) _init(GRE,groups); }

int CL_grep::_match(const char * SS, size_t& idx,
    size_t& len, int caseIgnore, CL<char> * groupToks)
{
  if (groupToks)  {
      assert(_groups);
      (void) groupToks->allClr();
      (void) groupToks->setMaxNodes();
      groupToks->setFlags(CL_DEL);
  }
  size_t SSlen = (SS? strlen(SS) : 0);
  if (parseTree.nodes() && SSlen)
    for (idx = 0; idx < SSlen; idx++)  {
      if (parseTree.match(&SS[idx],len,
          caseIgnore,groupToks))
          if (!EOL_anchor || idx + len >= SSlen)
              return 1;
      if (BOL_anchor)
          return 0;
    }
  return 0;
}

CL_grep::CL_grep(const char * GRE)
    : parseTree(), gre()
    , BOL_anchor(0), EOL_anchor(0), _groups(0)
{ (void) _init(GRE,0); }

int CL_grep::init(const char * GRE)
{ return _init(GRE); }

int CL_grep::match(const char * SS, size_t& idx,
        size_t& len, int caseIgnore)
{ return _match(SS,idx,len,caseIgnore,0); }

ostream& CL_grep::storeOn(ostream& os)
{ return gre.storeOn(os << _groups << endm); }

istream& CL_grep::readFrom(istream& is)
{
    int groups;
    (void) gre.readFrom(is >> groups >> nextm);
    (void) _init(gre.str(),groups);
    return is;
}

CL_ggrep::CL_ggrep(const char * GRE)
    : CL_grep(GRE,1), CL<char>(CL_DEL)
{ init(GRE); }

int CL_ggrep::init(const char * GRE)
{
    allClr();
    return CL_grep::_init(GRE,1);
}

int CL_ggrep::match(const char * SS, size_t& idx,
    size_t& len, int caseIgnore)
{
    return CL_grep::
        _match(SS,idx,len,caseIgnore,(CL<char> *)this);
}

istream& CL_ggrep::readFrom(istream& is)
{
    allClr();
    return CL_grep::readFrom(is);
}

grep_seq::grep_seq(const char * GRE, size_t GRE_len,
    size_t& tokLen, int groups)
    : grep_tok(), cstr()
{
    assert(GRE);
    size_t lastCspnIdx = tokLen = 0;
    size_t cspnLen;
    while (tokLen < GRE_len)  {
        if (!ncat(&GRE[tokLen],cspnLen
            = strcspn(&GRE[tokLen],"[()]*+?|\\.{}"),1)
            .ok())
            { tokLen = 0; return; }
        tokLen += cspnLen;
        if (tokLen >= GRE_len)
            break;
        if (strchr("{}",GRE[tokLen]))
            if (groups)
                break;  // "{}"
            else {
                ncat(&GRE[tokLen++],1);
                continue;
            }
        if (GRE[tokLen] != '\\')
            break;  // "[()]*+?|."
        lastCspnIdx = tokLen; // found '\\'
        if (++tokLen < GRE_len)
            ncat(&GRE[tokLen++],1);
        else {
            error(GRE,tokLen,"nothing following \\");
            tokLen = 0;
            return;
        }
    }
    if (tokLen < GRE_len
        && strspn(&GRE[tokLen],"*+?|"))  {
        if (lastCspnIdx == (--tokLen - 1))
            tokLen--;
        if (len())
            trunc(len()-1);
    }
    pack();
    if (!len())  {
        tokLen = 0;
    }
}

#pragma argsused
int grep_seq::matchOne
    (const char * SS, size_t& len_,
    int caseIgnore, CL<char> * groupToks)
{
    return (len_ = (ncmp(SS,len(),caseIgnore)?
        0 : len()));
}

grep_setTok::~grep_setTok() {}

#pragma argsused
int grep_setTok::match(char c, int caseIgnore)
{ assert(0); return 0; }

grep_setAlts::grep_setAlts(const char * SEQ)
    : grep_setTok(), cstr(SEQ)
{}

int grep_setAlts::match(char c, int caseIgnore)
{
    if (strchr(str(),c))
        return 1;
    if (caseIgnore)
        if (isupper(c))  {
            if (strchr(str(),::tolower(c)))
                return 1;
        }
        else if (strchr(str(),::toupper(c)))
            return 1;
    return 0;
}                  

grep_setRange::grep_setRange(int left, int right)
    : grep_setTok(), _left(left), _right(right)
{ assert(left < right); }

int grep_setRange::match(char c, int caseIgnore)
{
    if (caseIgnore)
        return (toupper(c) >= toupper(_left))
            && (toupper(c) <= toupper(_right));
    else
        return (c >= _left) && (c <= _right);
}

grep_set::grep_set(const char * GRE, size_t GRE_len,
     size_t& tokLen)
    : grep_tok(), CL<grep_setTok>(CL_DEL), negate(0)
{
  assert(GRE && GRE[0] == '[');
  tokLen = 1;
  if (GRE[1] == '^')  {
      negate = 1;
      tokLen++;
  }
  size_t cspnLen;
  cstr alts;
  while (tokLen < GRE_len)  {
    if (!alts.ncat(&GRE[tokLen],cspnLen
      = strcspn(&GRE[tokLen],"-]\\[()^*+?|.{}"),1)
      .ok())
      { tokLen = 0; return; }
    tokLen += cspnLen;
    if (tokLen >= GRE_len) {
      error(GRE,tokLen,"] not found");
      tokLen = 0; return; }
    if (strspn(&GRE[tokLen],"[()^*+?|.{}")) {
      error(GRE,tokLen,"illegal grep set char");
      tokLen = 0; return; }
    switch (GRE[tokLen])  {
      case '-':
        if (!alts.len()) {
          error(GRE,tokLen,"left range missing");
          tokLen = 0;
          return;
        }
        if (++tokLen >= GRE_len)  {
          error(GRE,tokLen,"right range missing");
          tokLen = 0;
          return;
        }
        if (GRE[tokLen] == '\\')  {
          if (++tokLen >= GRE_len)  {
            error(GRE,tokLen,
              "char following \\ missing");
            tokLen = 0;
            return;
          }
        }
        if (!insQDL(new grep_setRange
              (alts[alts.len()-1],GRE[tokLen++])))
           { tokLen = 0; return; }
        alts.rmv(alts.len()-1);
        if (!alts.len())
            break;
        // fall through for alts preceeding range
      case ']':
        if (alts.len())
            if (!insQDL(new grep_setAlts(alts.str())))
                { tokLen = 0; return; }
        tokLen++;
        alts.clr();
        if (!nodes()) {
          error(GRE,tokLen,"nothing in set");
          tokLen = 0;
          return;
        }
        return;
      case '\\':
        if (++tokLen < GRE_len)
          alts.ncat(&GRE[tokLen],1);
        else {
          error(GRE,tokLen,"nothing following \\");
          tokLen = 0;
          return;
        }
        break;
      default: assert(0); break;
    }
  }
  tokLen = 0;
}

#pragma argsused
int grep_set::matchOne
    (const char * SS, size_t& len,
    int caseIgnore, CL<char> * groupToks)
{  // No group tokens will be matched within a set!
    if (negate)  {
      for (unsigned i = 0; i < nodes(); i++)
        if (!atGet(i)->match(*SS,caseIgnore))
            return len = 1;
    }
    else
      for (unsigned i = 0; i < nodes(); i++)
        if (atGet(i)->match(*SS,caseIgnore))
            return len = 1;
    return 0;
}

grep_alt::grep_alt(const char * GRE, size_t GRE_len,
    grep_tok * leftTok, size_t& rightTokLen,
    int groups)
    : _left(leftTok), _right(0)
{
    assert(leftTok);
    _right = grep_tok::newTok
        (GRE,GRE_len,rightTokLen,groups);
}

grep_alt::~grep_alt()
{
    delete _left;
    delete _right;
}

int grep_alt::matchOne(const char * SS,
    size_t& len, int caseIgnore, CL<char> * groupToks)
{   // count is always One for grep_alt!
    assert(_left && _right);
    if (_left->matchOne(SS,len,caseIgnore,groupToks))
        return 1;
    if (_right->matchOne(SS,len,caseIgnore,groupToks))
        return 1;
    return 0;
}


// To compile demo: bcc -ml clgrep cstr cl

//#define CLGREP_DEMO
#ifdef CLGREP_DEMO

char input[] = "$not 12noway. .wayne o_kay //";
CL_grep idsOnly("[ ]([a-z]|[A-Z_])+[0-9a-zA-Z_]*[ ]");
CL_grep noway(".*now.y");
CL_grep way("way[^\\.]");
CL_grep not("^$not[ ]+");
CL_grep line("^.*$");
CL_grep comments("((/\\*)|(//))");
CL_ggrep wayne("\\.{[a-z]+} ");

void GREP(CL_grep * grep, const char * input)
{
    size_t idx, len;
    assert(grep && input);
    grep->printOn(cout << "\nRE:    ") << endl;
    cout << "Input: " << input << endl;
    cout << "Found: ";
    if (grep->match(input,idx,len))  {
        cstr carrots(' ',idx,len);
        carrots.ncat('^',len);
        carrots.printOn(cout);
    }
    cout << endl;
}

main()
{
    GREP(&idsOnly,input);
    GREP(&noway,input);
    GREP(&way,input);
    GREP(&not,input);
    GREP(&line,input);
    GREP(&comments,input);
    GREP(&wayne,input);
    if (wayne.groupToksOk() && wayne.nodes())
        cout << "Wayne token: " << wayne.top() << endl;
    return 0;
}

#endif
